unit RyMenus;

interface

{.$DEFINE RY}

uses Windows,SysUtils,Classes,Graphics,ImgList,Menus,Forms,
     Commctrl;

type
  TRyMenu = class(TObject)
  private
    FFont: TFont;
    FGutterColor: TColor;
    FMenuColor: TColor;
    FSelectedColor: TColor;
    FSelLightColor: TColor;
    procedure SetFont(const Value: TFont);
    procedure SetSelectedColor(const Value: TColor);
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(Menu: TMenu; Item: TMenuItem);
    procedure MeasureItem(Sender: TObject; ACanvas: TCanvas;
              var Width, Height: Integer);
    procedure AdvancedDrawItem(Sender: TObject; ACanvas: TCanvas;
              ARect: TRect; State: TOwnerDrawState);
    property MenuColor: TColor read FMenuColor write FMenuColor;
    property GutterColor: TColor read FGutterColor write FGutterColor;
    property SelectedColor: TColor read FSelectedColor write SetSelectedColor;
    property Font: TFont read FFont write SetFont; // .  
  end;

var
  RyMenu: TRyMenu;// . -  TRyMenu
                  //.   .
  clf,cls,cpf,cpc:tcolor;
  logo:string;
  _h,_cnt:smallint;
  //Res: TResourceStream;

implementation
//{$R skin/plugins/fnt_res/rfont.res}//  -

{$IFDEF RY}
uses RyUtils, RyClassesUtils;
{$ELSE}
type
  TRGB = packed record
    R, G, B: Byte;
  end;
{$ENDIF RY}

const _DefSelColor: TColor = $00D7AFAF;{ $00FFAEB0;}// 
      {_DefSelLightColor: TColor = $00DECDCD;}
var
  FMonoBitmap : TBitmap;
  BmpCheck: array[Boolean] of TBitmap; // bmp 16x16    

{$IFNDEF RY}
function Max(A,B: Integer): Integer;
begin
 if A < B then Result:= B
  else Result:= A
end;

//   
function GetRGB(const Color: TColor): TRGB;
var iColor: TColor;
begin
 iColor  := ColorToRGB(Color);
 Result.R:= GetRValue(iColor);
 Result.G:= GetGValue(iColor);
 Result.B:= GetBValue(iColor);
end;

//  2-  
function GetHighlightColor(const Color: TColor) : TColor;
var fFrom,fTo: TRGB;
begin
 FFrom:= GetRGB(Color);
 FTo  := GetRGB(clWhite);

 Result:= RGB(
  Round(FFrom.R + (FTo.R - FFrom.R) * 0.5),
  Round(FFrom.G + (FTo.G - FFrom.G) * 0.5),
  Round(FFrom.B + (FTo.B - FFrom.B) * 0.5))
end;
{$ENDIF RY}

{ TRyMenuItem }
constructor TRyMenu.Create;
begin
 _h:=16;                   //.
 logo:=' R A M E D I A    T V';//
 FGutterColor:= clskyblue; //.
 FMenuColor  := clwhite;   //.
 clf:=clblack;             //.
 cls:=clblue;              //..
 //
 cpf:=clwhite;             //.
 cpc:=clblue;              //.
 //
 FSelectedColor:= _DefSelColor;
 FSelLightColor:= GetHighlightColor(_DefSelColor);
 FFont:= TFont.Create;
 Font := Screen.MenuFont;// .
end;

destructor TRyMenu.Destroy;
begin
 FFont.Free;
 inherited;
end;

procedure TRyMenu.Add(Menu: TMenu; Item: TMenuItem);

 procedure InitItem(Item: TMenuItem);
 begin
  Item.OnAdvancedDrawItem:= Self.AdvancedDrawItem;
  if not (Item.GetParentComponent is TMainMenu) then
   Item.OnMeasureItem    := Self.MeasureItem;
 end;

 procedure InitItems(Item: TMenuItem);
 //   + 
 var I: Word;
 begin
  I:= 0;
  while I < Item.Count do
   begin
    InitItem(Item[I]);
    if Item[I].Count > 0 then
     InitItems(Item[I]);
    Inc(I);
   end;
 end;

begin
 if Assigned(Menu) then
  begin
   _cnt:=menu.Items.Count-2;
   InitItems(Menu.Items);
   Menu.OwnerDraw:= True;
  end;
 if Assigned(Item) then
  begin
   InitItem(Item);
   InitItems(Item);
  end;
end;

//-
procedure TRyMenu.AdvancedDrawItem(Sender: TObject; ACanvas: TCanvas;
          ARect: TRect; State: TOwnerDrawState);
const
 ROP_DSPDxax = $00E20746;
const
 // 
 _Flags: LongInt = DT_NOCLIP or DT_VCENTER or DT_END_ELLIPSIS or DT_SINGLELINE;
 _FlagsTopLevel: array[Boolean] of Longint = (DT_LEFT, DT_CENTER);
 _FlagsShortCut: {array[Boolean] of} Longint = (DT_RIGHT);
 _RectEl: array[Boolean] of Byte = (0, 6);//.
var
 TopLevel : Boolean;
 Gutter   : Word; //.
 ImageList: TCustomImageList;
 // -
 x1,x2,y1,y2,h: integer;
 LogFont  : TLogFont;
 tmpRect  : TRect;
begin
 //
 with TMenuItem(Sender),ACanvas do
  begin
   if tag=100 then h:=_h
    else h:=0;//  -

   // -
   GetObject(acanvas.Font.Handle, SizeOf(LogFont), @LogFont);
   with LogFont do
    begin
     lfEscapement   := 90 * 10;
     lfOrientation  := 90 * 10;
     lfOutPrecision := OUT_TT_ONLY_PRECIS;
     lfFaceName     := 'Tahoma';
     lfHeight       := 12;
     lfWeight       := FW_BOLD;
     lfQuality      := PROOF_QUALITY;
    end;
   with aCanvas do
    begin
     Font.Handle:= CreateFontIndirect(LogFont);
     Font.Color := cpf;
     Brush.Color:= cpc;
    end;
   x1:= GetSystemMetrics(SM_CXEDGE)-  GetSystemMetrics(SM_CXBORDER)-2;
   x2:= h-2;
   y1:= GetSystemMetrics(SM_CYCAPTION) - GetSystemMetrics(SM_CYEDGE) -
         GetSystemMetrics(SM_CYBORDER) - 25;
   y2:= _cnt*(arect.Bottom-arect.Top)+GetSystemMetrics(SM_CYEDGE)+
        GetSystemMetrics(SM_CYBORDER)+2;
   tmpRect:= Rect(x1, y1, x2, y2);
   aCanvas.FillRect(tmpRect);
   DrawText(aCanvas.Handle,PChar(logo),-1,tmpRect,DT_BOTTOM or DT_SINGLELINE);
   //
   TopLevel := GetParentComponent is TMainMenu;
   ImageList:= GetImageList; //  ImageList
   Font:= FFont;             // 
   if Assigned(ImageList) then
    Gutter:= ImageList.Width + 9   //4- pix   +  + 5- 
   else
   if IsLine then
    Gutter:= Max(TextHeight('W'), 18) + 4
   else
    Gutter:= ARect.Bottom - ARect.Top + 4;// =  + 2 + 2 pix

   Pen.Color:= cls;
   if (odSelected in State) then   //  
    begin
     Brush.Color:= SelectedColor;  // -
     Rectangle(ARect.Left+h, ARect.Top, ARect.Right, ARect.Bottom);
    end else
   if TopLevel then                //   .
    begin
     if (odHotLight in State) then //   
      begin
       Pen.Color:= clBtnShadow;
       Brush.Color:= FSelectedColor;
       Rectangle(ARect.Left+h, ARect.Top, ARect.Right, ARect.Bottom);
      end else begin
       Brush.Color:= clBtnFace;
       FillRect(ARect);
      end
    end else begin                 // -
      Brush.Color:= GutterColor;   //
      FillRect(Rect(ARect.Left+h, ARect.Top, Gutter+h, ARect.Bottom));
      Brush.Color:= MenuColor;
      FillRect(Rect(Gutter+h, ARect.Top, ARect.Right+h, ARect.Bottom));
     end;
   if Checked then
    begin                          //. 
     Brush.Color:= FSelLightColor;
     if Assigned(ImageList) and (ImageIndex > -1) then
      // ,   
      RoundRect(ARect.Left + 2+h, ARect.Top, Gutter - 2 - 1+h,
       ARect.Bottom, _RectEl[RadioItem], _RectEl[RadioItem])
      else // 
       Draw((ARect.Left + 2 + Gutter - 1 - 2 - BmpCheck[RadioItem].Width) div 2+h,
        (ARect.Top + ARect.Bottom - BmpCheck[RadioItem].Height) div 2,
        BmpCheck[RadioItem])
    end;

   if Assigned(ImageList) and ((ImageIndex > -1) and (not TopLevel)) then
    if Enabled then
     ImageList.Draw(ACanvas, ARect.Left + 4+h,
      (ARect.Top + ARect.Bottom - ImageList.Height) div 2,
      ImageIndex, True) // .
    else begin          //  
     //
     with FMonoBitmap do
      begin
       Width := ImageList.Width;
       Height:= ImageList.Height;
       Canvas.Brush.Color := clWhite;
       Canvas.FillRect(Rect(h, 0, Width+h, Height));
       ImageList_DrawEx(ImageList.Handle,ImageIndex,
        Canvas.Handle,0,0,0,0,CLR_DEFAULT,0,ILD_NORMAL);
      end;
     Brush.Color:= clBtnShadow;
     Windows.SetTextColor(Handle, clWhite);
     Windows.SetBkColor(Handle, clBlack);
     BitBlt(Handle,ARect.Left+4+h,
      (ARect.Top+ARect.Bottom-ImageList.Height) div 2,
      FMonoBitmap.Width,FMonoBitmap.Height,
      FMonoBitmap.Canvas.Handle,0,0,ROP_DSPDxax);
     //
    end;
   with Font do
    begin
     if (odDefault in State)  then Style:= [fsBold];
     if (odDisabled in State) then Color:= clGray
      else Color:= clf;
    end;

   Brush.Style:= bsClear;
   if TopLevel then //  
    else Inc(ARect.Left, Gutter + 5+h);

   if IsLine then   // 
    begin
     Pen.Color:= clBtnShadow;
     MoveTo(ARect.Left ,ARect.Top + (ARect.Bottom - ARect.Top) div 2);
     LineTo(ARect.Right,ARect.Top + (ARect.Bottom - ARect.Top) div 2);
    end else
    begin           // 
     Windows.DrawText(Handle, PChar(Caption), Length(Caption), ARect,
      _Flags or _FlagsTopLevel[TopLevel]);
     if ShortCut <> 0 then begin
       Dec(ARect.Right, 5);
       Windows.DrawText(Handle, PChar(ShortCutToText(ShortCut)),
        Length(ShortCutToText(ShortCut)), ARect,
        _Flags or _FlagsShortCut);
      end
    end
  end
end;

// -
procedure TRyMenu.MeasureItem(Sender: TObject; ACanvas: TCanvas;
          var Width, Height: Integer);
var ImageList: TCustomImageList;
begin
 with TMenuItem(Sender) do
  begin
   ImageList   := GetImageList;
   ACanvas.Font:= FFont; // -
   if Assigned(ImageList) then
    begin
     if IsLine then
      if ImageList.Height > 20 then //> 20  
       Height:= 11 else Height := 5
     else with ACanvas do
      begin
       Width:= ImageList.Width;
       if Width < 8 then Width := 16 else Width := Width + 8;
       Width := Width + TextWidth(Caption + ShortCutToText(ShortCut)) + 15+20;
       Height:= Max(ACanvas.TextHeight('W'), ImageList.Height);
       if Height < 14 then Height := 18 else Height := Height + 4;
      end
    end else with ACanvas do
     begin
      Height := Max(TextHeight('W'), 18);
      if IsLine then
       if Height > 20 then //> 20  
        Height:= 11 else Height:= 5;
        Width := 20+15+20+TextWidth(Caption + ShortCutToText(ShortCut));
     end
  end
end;

procedure TRyMenu.SetFont(const Value: TFont);
begin
 FFont.Assign(Value);
end;

procedure InitBmp(Bmp: TBitmap; Radio: Boolean);
begin
 with Bmp, Canvas do
  begin
   Width := 14; Height:= 14;
   Pen.Color  := clBlack;
   Brush.Color:= GetHighlightColor(_DefSelColor);
   Rectangle(0, 0, Width, Height);
   Brush.Color:= clBlack;
   if Radio then Ellipse(4, 4, Width - 1 - 3, Height - 1 - 3)
    else Rectangle(4, 4, Width - 1 - 3, Height - 1 - 3);
  end
end;

procedure TRyMenu.SetSelectedColor(const Value: TColor);
begin
 FSelectedColor:= Value;
 FSelLightColor:= GetHighlightColor(Value);
 {_DefSelColor := Value;          //. bmp
 InitBmp(BmpCheck[False], False); //  .
 InitBmp(BmpCheck[True], True);}
end;

initialization
 {try
 Res := TResourceStream.Create(hInstance, 'MY_FONT', Pchar('rfnt'));
 Res.SavetoFile('raFNT.ttf');
 Res.Free;
 AddFontResource(PChar('skin\plugins\raf.ttf'));
 SendMessage(HWND_BROADCAST,$001D,0,0);
 except end;}
 //
 BmpCheck[False]:= TBitmap.Create;
 BmpCheck[True]:= TBitmap.Create;
 InitBmp(BmpCheck[False], False);
 InitBmp(BmpCheck[True], True);
 FMonoBitmap := TBitmap.Create;
 FMonoBitmap.Monochrome := True;
 RyMenu:= TRyMenu.Create;

finalization
 FMonoBitmap.Free;
 BmpCheck[False].Free;
 BmpCheck[True].Free;
 RyMenu.Free;
 //
 //RemoveFontResource(PChar('raf.ttf'));
 //SendMessage(HWND_BROADCAST,$001D,0,0);
 
end.
